@ECHO OFF
REM ============================================================
REM wrapper for the SDCC-Compiler
REM
REM usage: sdcc-compile [-o directory] [-s directory] [--debug]
REM                     [--prof] [-code bank] cFileBaseName
REM 
REM   -o       specifies the directory for the object files
REM   -s       specifies the directory of the source file
REM   --debug  tells that intermediate files are kept for debugging
REM   --prof   adds prologue and epilogue code for profiling
REM   --code   specifies the bank number where the code should later be
REM            linked to
REM
REM   external environment variables
REM       GAWK_PROGRAM:             path of executable for gawk
REM       MV_PROGRAM:               path of executable for mv
REM       KEEP_LIST_FILES:          tells that assembly list files are kept
REM       RM_PROGRAM:               path of executable for rm
REM       TEST_PROGRAM:             path of executable for test
REM       SDCC_PATH:                root directory of SDCC compiler
REM       SDCC_USER_DEFINES:        options for SDCC compiler
REM       TARGET_INCLUDE_DIR:       include directory for current compilation
REM       TARGET_INCLUDE_ROOT_DIR:  include directory for current compilation
REM       TOOLS_DIR:                directory for tools and gawk scripts
REM
REM ============================================================

SETLOCAL

REM -- set variables --
SET sdccIncludeOptions=-I%TARGET_INCLUDE_ROOT_DIR%
SET sdccIncludeOptions=%sdccIncludeOptions% -I%TARGET_INCLUDE_DIR%
SET sdccIncludeOptions=%sdccIncludeOptions% -I%SDCC_PATH%\include

SET sdccDefines=-DSDCC_PORT=gbz80 -DSDCC_MODEL_SMALL -DGB -D__gbz80
SET sdccDefines=%sdccDefines% -D_gbz80 -Dgbz80  -DGAMEBOY -DINT_16_BITS
SET sdccDefines=%sdccDefines% -DBUGFIX

REM -- add user specific definitions --
SET sdccDefines=%sdccDefines% %SDCC_USER_DEFINES%

REM -- set compilation options --
SET sdccOptions=-S -lang-c99 -mgbz80
SET sdccOptions=%sdccOptions% --debug
SET sdccOptions=%sdccOptions% --i-code-in-asm  --opt-code-size

REM -- skip certain warnings --
REM    18  different levels of indirection
REM    85  unreferenced function argument
REM   126  unreachable code (optimized away)
SET sdccOptions=%sdccOptions% --disable-warning 18
SET sdccOptions=%sdccOptions% --disable-warning 85
SET sdccOptions=%sdccOptions% --disable-warning 126

REM -- remove some optimizations --
SET sdccOptions=%sdccOptions% --nogcse

SET sdccObjectDirectory=.
SET sdccSourceDirectory=.
SET sdccCodeSegment=--codeseg BASE --constseg BASE
SET sdccIsDebugging=false
SET sdccFileName=???

REM -- parse options --
:REPEAT
  IF %1x==x GOTO :DONE
  IF NOT %1==-o GOTO :ELSE11
    SHIFT
    SET sdccObjectDirectory=%1
    GOTO :ENDIF1
  :ELSE11
  IF NOT %1==-s GOTO :ELSE12
    SHIFT
    SET sdccSourceDirectory=%1
    GOTO :ENDIF1
  :ELSE12
  IF NOT %1==--debug GOTO :ELSE13
    SET sdccIsDebugging=true
    GOTO :ENDIF1
  :ELSE13
  IF NOT %1==--prof GOTO :ELSE14
    SET sdccOptions=%sdccOptions% --profile
    GOTO :ENDIF1
  :ELSE14
  IF NOT %1==--code GOTO :ELSE15
    SHIFT
    SET sdccCodeSegment=--codeseg CODE_%1 --constseg CODE_%1
    GOTO :ENDIF1
  :ELSE15
    SET sdccFileName=%1
  :ENDIF1
  SHIFT
  GOTO :REPEAT
:DONE

SET sdccObjectFileName=%sdccObjectDirectory%\%sdccFileName%.o
SET sdccSourceFileName=%sdccSourceDirectory%\%sdccFileName%.c
SET sdccOptions=%sdccOptions% %sdccCodeSegment%

REM -- test for existence of target and whether it is older than source
%TEST_PROGRAM% -e %sdccObjectFileName% -a %sdccObjectFileName% -nt %sdccSourceFileName%
IF %ERRORLEVEL%==1 GOTO :ELSE2
  SET updateIsNecessary=false
  GOTO :ENDIF2
:ELSE2
  SET updateIsNecessary=true
:ENDIF2

REM IF EXIST %sdccObjectFileName% GOTO :ELSE2
REM   SET updateIsNecessary=true
REM   GOTO :ENDIF2
REM :ELSE2
REM   IF %sdccObjectFileName% GEQ %sdccSourceFileName% GOTO :ELSE21
REM     SET updateIsNecessary=true
REM     GOTO :ENDIF21
REM   :ELSE21
REM     SET updateIsNecessary=false
REM   :ENDIF21
REM :ENDIF2

IF %updateIsNecessary%==true GOTO :ELSE3
  ECHO === %sdccFileName%.o is current ===
  GOTO :ENDIF3
:ELSE3
  REM -- do the compilation and assembly --
  SET sdccAllOptions=%sdccIncludeOptions% %sdccDefines% %sdccOptions%

  ECHO === Compile %sdccFileName%.c ===
  REM suppress the "from type ... to type ..." message in SDCC
  SET tempFile=temp
  %SDCC_PATH%\BIN\SDCC %sdccAllOptions% %sdccSourceFileName% 2>%tempFile%

  IF ERRORLEVEL 1 GOTO :ELSE31
    SET compilationResult=0
    GOTO :ENDIF31
  :ELSE31
    SET compilationResult=1
  :ENDIF31
  
  FINDSTR /V /C:"from type " <%tempFile% | FINDSTR /V /C:"to type "
  %RM_PROGRAM% %tempFile%

  IF %compilationResult%==0 GOTO :ELSE32
    ECHO === ### compilation failed ### ===
    GOTO :DONE
    GOTO :ENDIF32
  :ELSE32
    ECHO === Assemble %sdccFileName%.asm ===
    %GAWK_PROGRAM% -f %TOOLS_DIR%\rmAsmJunk.awk %sdccFileName%.asm  >%sdccFileName%.xxx
    %SDCC_PATH%\BIN\AS-GBZ80 -pogl %sdccFileName%.o %sdccFileName%.xxx

    %MV_PROGRAM% %sdccFileName%.o %sdccObjectDirectory%
    %RM_PROGRAM% %sdccFileName%.i

    IF %KEEP_LIST_FILES%==false GOTO :ELSE321
      %MV_PROGRAM% %sdccFileName%.lst %sdccObjectDirectory%
      GOTO :ENDIF32
    :ELSE321
      %RM_PROGRAM% %sdccObjectDirectory%\%sdccFileName%.lst
    :ENDIF321

    %RM_PROGRAM% %sdccFileName%.map %sdccFileName%.ihx
  :ENDIF32
:ENDIF3


REM --- clean up ---
%RM_PROGRAM% %sdccObjectDirectory%\%sdccFileName%.adb
%RM_PROGRAM% %sdccObjectDirectory%\%sdccFileName%.asm
%RM_PROGRAM% %sdccObjectDirectory%\%sdccFileName%.lst
%RM_PROGRAM% %sdccObjectDirectory%\%sdccFileName%.xxx

IF %sdccIsDebugging%==true GOTO :ELSE4
  %RM_PROGRAM% %sdccFileName%.adb %sdccFileName%.asm
  %RM_PROGRAM% %sdccFileName%.lst %sdccFileName%.xxx
  GOTO :ENDIF4
:ELSE4
  %MV_PROGRAM% %sdccFileName%.adb %sdccObjectDirectory%
  %MV_PROGRAM% %sdccFileName%.asm %sdccObjectDirectory%
  %MV_PROGRAM% %sdccFileName%.lst %sdccObjectDirectory%
  %MV_PROGRAM% %sdccFileName%.xxx %sdccObjectDirectory%
:ENDIF4

:DONE

ENDLOCAL
